class Placeholder:
    """
        Base class for placeholders.

        :var str full_name:
        :var list arguments: Arguments for placeholder

    """
    def __init__(self, full_name, arguments):
        self.full_name = full_name
        self.arguments = arguments

    def exec(self):
        """Each placeholder must have exec function"""
        pass

    def __str__(self):
        return f"{self.full_name} {self.arguments}"


class FuncPlaceholder(Placeholder):
    """
        Represents placeholder which takes function to execute later.
        Returned value is fully determined by func itself.

        :var func func: Function to execute. Takes provided arguments from placeholder.
        :var list args: Parsed arguments as list
    """
    def __init__(self, full_name, args, func):
        super().__init__(full_name, args)
        self.func = func

    def exec(self):
        """Executes func and return its returned value"""
        return self.func(self.arguments)


class FuncStrPlaceholder(FuncPlaceholder):
    """
            Represents placeholder which takes function to execute later.
            The returned value is enclosed with double quotes to denote JSON string value.

            :var func func: Function to execute. Takes provided arguments from placeholder.
        """
    def __init__(self, full_name, args, func):
        super().__init__(full_name, args, func)

    def exec(self):
        """Executes func and return its value enclosed as string"""
        return f'"{self.func(self.arguments)}"'



from jdog.placeholder.placeholder import Placeholder


class RangePlaceholder(Placeholder):
    def __init__(self, full_name, args):
        super().__init__(full_name, args)
        self.prop = args[0]
        self.low = args[1]
        if len(args) > 2:
            self.high = args[2]

    def exec(self):
        pass




import re
import json
import random

from faker import Faker

from jdog.node import PropertyNode, ScalarNode, PlaceholderNode, ObjectNode, ArrayNode, RangeNode
from jdog.placeholder.name import NamePlaceholder, NamePlaceholderOption
from jdog.placeholder.option import OptionPlaceholder
from jdog.placeholder.placeholder import FuncPlaceholder, FuncStrPlaceholder
from jdog.placeholder.range import RangePlaceholder


class NoMatchingPlaceholder(Exception):
    """Raised when no matching placeholder was found."""

    def __init__(self, token):
        super().__init__(f'No matching placeholder for {token} found.')
        self.token = token


class SchemeParser:
    """
        Parsing provided scheme and construct node structure for later data generation.
        Defines predefined placeholders.

        :var faker faker: Faker instance with provided language
        :var boolean strict: If parser should raise :class:`.NoMatchingPlaceholder`
    """

    NAME = 'name'
    FIRST_NAME = 'first_name'
    LAST_NAME = 'last_name'
    CITY = 'city'
    STREET_ADDRESS = 'street_address'
    AGE = 'age'
    NUMBER = 'number'
    LOREM = 'lorem'
    OPTION = 'option'
    EMPTY = 'empty'
    RANGE = 'range'
    BOOL = 'bool'

    def __init__(self, lang='en-US', strict=False):
        self.faker = Faker(lang)
        self.strict = strict
        self.compiled_matchers = {
            SchemeParser.NAME: re.compile(r'^{{name\(?([f,m]?)\)?}}$'),
            SchemeParser.FIRST_NAME: re.compile(r'^{{first_name\(?([f,m]?)\)?}}$'),
            SchemeParser.LAST_NAME: re.compile(r'^{{last_name\(?([f,m]?)\)?}}$'),
            SchemeParser.CITY: re.compile('^{{city}}$'),
            SchemeParser.STREET_ADDRESS: re.compile('^{{street_address}}$'),
            SchemeParser.AGE: re.compile('^{{age}}$'),
            SchemeParser.NUMBER: re.compile(r'^{{number\((.*)\)}}$'),
            SchemeParser.LOREM: re.compile(r'^{{lorem\((.*)\)}}$'),
            SchemeParser.OPTION: re.compile(r'^{{option\((.*)\)}}$'),
            SchemeParser.EMPTY: re.compile(r'^{{empty}}$'),
            SchemeParser.RANGE: re.compile(r'^{{range\((.*)\)}}$'),
            SchemeParser.BOOL: re.compile(r'^{{bool}}$')
        }
        self.matchers = {
            SchemeParser.NAME: lambda token: self.compiled_matchers[SchemeParser.NAME].match(token),
            SchemeParser.FIRST_NAME: lambda token: self.compiled_matchers[SchemeParser.FIRST_NAME].match(token),
            SchemeParser.LAST_NAME: lambda token: self.compiled_matchers[SchemeParser.LAST_NAME].match(token),
            SchemeParser.CITY: lambda token: self.compiled_matchers[SchemeParser.CITY].match(token),
            SchemeParser.STREET_ADDRESS: lambda token: self.compiled_matchers[SchemeParser.STREET_ADDRESS].match(token),
            SchemeParser.AGE: lambda token: self.compiled_matchers[SchemeParser.AGE].match(token),
            SchemeParser.NUMBER: lambda token: self.compiled_matchers[SchemeParser.NUMBER].match(token),
            SchemeParser.LOREM: lambda token: self.compiled_matchers[SchemeParser.LOREM].match(token),
            SchemeParser.OPTION: lambda token: self.compiled_matchers[SchemeParser.OPTION].match(token),
            SchemeParser.EMPTY: lambda token: self.compiled_matchers[SchemeParser.EMPTY].match(token),
            SchemeParser.RANGE: lambda token: self.compiled_matchers[SchemeParser.RANGE].match(token),
            SchemeParser.BOOL: lambda token: self.compiled_matchers[SchemeParser.BOOL].match(token),
        }
        self.placeholders = {
            SchemeParser.NAME:
                lambda token, args: self._placeholder_name(token, args),
            SchemeParser.FIRST_NAME:
                lambda token, args: self._placeholder_first_name(token, args),
            SchemeParser.LAST_NAME:
                lambda token, args: self._placeholder_last_name(token, args),
            SchemeParser.CITY:
                lambda token, _: FuncStrPlaceholder(token, _, lambda __: self.faker.city()),
            SchemeParser.STREET_ADDRESS:
                lambda token, _: FuncStrPlaceholder(token, _, lambda __: self.faker.street_address()),
            SchemeParser.AGE:
                lambda token, _: FuncPlaceholder(token, _, lambda __: random.randint(1, 99)),
            SchemeParser.NUMBER:
                lambda token, args: FuncPlaceholder(token, args, lambda xargs: random.randint(int(xargs[0]), int(xargs[1]) - 1)),
            SchemeParser.LOREM:
                lambda token, args: FuncStrPlaceholder(token, args, lambda xargs: self.faker.sentence(nb_words=int(xargs[0]))),
            SchemeParser.OPTION:
                lambda token, args: OptionPlaceholder(token, args),
            SchemeParser.EMPTY:
                lambda token, _: FuncStrPlaceholder(token, _, lambda __: ''),
            SchemeParser.RANGE:
                lambda token, args: RangePlaceholder(token, args),
            SchemeParser.BOOL:
                lambda token, _: FuncStrPlaceholder(token, _, lambda __: str(random.random() > 0.5).lower())
        }

    def _placeholder_name(self, token, args):
        if args and args[0] == 'm':
            return NamePlaceholder(token, self.faker, NamePlaceholderOption.GENDER_MALE)
        elif args and args[0] == 'f':
            return NamePlaceholder(token, self.faker, NamePlaceholderOption.GENDER_FEMALE)

        return NamePlaceholder(token, self.faker)

    def _placeholder_first_name(self, token, args):
        if args and args[0] == 'm':
            return NamePlaceholder(token, self.faker,
                                   option=NamePlaceholderOption.FIRST_NAME | NamePlaceholderOption.GENDER_MALE)
        elif args and args[0] == 'f':
            return NamePlaceholder(token, self.faker,
                                   option=NamePlaceholderOption.FIRST_NAME | NamePlaceholderOption.GENDER_FEMALE)
        return NamePlaceholder(token, self.faker, option=NamePlaceholderOption.FIRST_NAME)

    def _placeholder_last_name(self, token, args):
        if args and args[0] == 'm':
            return NamePlaceholder(token, self.faker,
                                   option=NamePlaceholderOption.LAST_NAME | NamePlaceholderOption.GENDER_MALE)
        elif args and args[0] == 'f':
            return NamePlaceholder(token, self.faker,
                                   option=NamePlaceholderOption.LAST_NAME | NamePlaceholderOption.GENDER_FEMALE)
        return NamePlaceholder(token, self.faker, option=NamePlaceholderOption.LAST_NAME)

    @staticmethod
    def _parse_arguments(p):
        pattern = r'([^,]+\(.+?\)}})|([^,]+)'
        matches = re.findall(pattern, p)

        return [x[0] if x[0] != '' else x[1] for x in matches]

    @staticmethod
    def _is_like_token(arg):
        return re.match(r'^{{?.*}?}$', arg)

    def _match_token(self, token):
        for key in self.matchers:
            m = self.matchers[key](str(token))

            if m is None:
                continue

            args = []
            if len(m.groups()) > 0:
                parsed_args = self._parse_arguments(m.group(1))
                # parse args to placeholders
                for arg in parsed_args:
                    if self._is_like_token(arg):
                        args.append(self._sub_parse(arg))
                    else:
                        args.append(arg)

            return self.placeholders[key](token, args)

        if self.strict and self._is_like_token(token):
            raise NoMatchingPlaceholder(token)
        else:
            return None

    def add_matcher(self, key, f_matcher, f_placeholder):
        """
            Add new matcher for returning new placeholder

        :param key: Unique matcher key. If provided existing one, old behavior is replaced.
        :param f_matcher: Function which takes one str argument.
            Should return re.Match object or None if no match found.
        :param f_placeholder: Function which takes accepted token and its parsed arguments - if present.
            Should return one of :class:`~jdog.placeholder.placeholder.Placeholder` object.
        """

        self.matchers[key] = f_matcher
        self.placeholders[key] = f_placeholder

    def placeholder_keys(self):
        """
            Defined placeholder keys

        :return: Defined placeholder keys
        :rtype: list
        """
        return self.placeholders.keys()

    def _sub_parse(self, sub):
        if isinstance(sub, list):
            array = []
            for x in sub:
                array.append(self._sub_parse(x))
            return ArrayNode(array)
        elif isinstance(sub, dict):
            props = []
            for key in sub:
                # Key could be token
                placeholder = self._match_token(key)
                if placeholder is None:
                    props.append(PropertyNode(ScalarNode(key), self._sub_parse(sub[key])))
                else:
                    # special case for RangePlaceholder
                    if isinstance(placeholder, RangePlaceholder):
                        props.append(RangeNode(placeholder.prop, placeholder.low, self._sub_parse(sub[key]),
                                               placeholder.high if hasattr(placeholder, 'high') else None))
                    else:
                        props.append(PropertyNode(PlaceholderNode(placeholder), self._sub_parse(sub[key])))
            return ObjectNode(props)
        else:
            placeholder = self._match_token(sub)
            if placeholder is None:
                return ScalarNode(sub)
            else:
                return PlaceholderNode(placeholder)

    def parse(self, scheme):
        """
            Parse given scheme and return node structure representation of the scheme.

        :param str scheme: Scheme to parse
        :return: Node structure representing current scheme
        """
        js = json.loads(scheme)
        return self._sub_parse(js)




import random
from jdog.placeholder.placeholder import Placeholder


class OptionPlaceholder(Placeholder):
    """
        Option placeholder picks randomly one of arguments.
        Each argument can be nested placeholder.
    """
    def __init__(self, full_name, args):
        super().__init__(full_name, args)

    def exec(self):
        """
            Choose randomly one of arguments. If chosen argument is placeholder executes it
            and return otherwise return argument enclosed as string
        """
        pick = random.choice(self.arguments)
        try:
            return pick.exec()
        except AttributeError:
            return f'"{pick}"'



from enum import IntFlag
from jdog.placeholder.faker import FakerPlaceholder


class NamePlaceholderOption(IntFlag):
    FULL_NAME = 1
    FIRST_NAME = 2
    LAST_NAME = 4
    GENDER_BOTH = 8
    GENDER_MALE = 16
    GENDER_FEMALE = 32


class NamePlaceholder(FakerPlaceholder):
    """Generates random name, with specified parameter it can be set to generate specific type of name"""

    def __init__(self, full_name, fake_provider, option=NamePlaceholderOption.FULL_NAME):
        """
        Parameters
        ----------
        full_name : str
            The full name, non parsed of placeholder
        fake_provider: Faker
            Faker object to generate name
        option: NamePlaceholderOption
            Indicates which kind of name should be generated and if gender matters
            Possible values are FULL_NAME, FIRST_NAME, LAST_NAME
            and for gender  GENDER_BOTH, GENDER_MALE or GENDER_FEMALE
            Combining first and latter is accepted. E.g. FULL_NAME + GENDER_MALE will
            generate male full names.
            FULL_NAME and GENDER_BOTH are default
        """
        super().__init__(full_name, [], fake_provider)
        self.option = option

    def exec(self):
        """
        :return: Name based on provided options.
        :rtype string
        """
        if NamePlaceholderOption.FIRST_NAME in self.option:
            return f'"{self._first_name()}"'
        elif NamePlaceholderOption.LAST_NAME in self.option:
            return f'"{self._last_name()}"'
        else:
            return f'"{self._name()}"'

    def _name(self):
        if NamePlaceholderOption.GENDER_MALE in self.option:
            return self.faker.name_male()
        elif NamePlaceholderOption.GENDER_FEMALE in self.option:
            return self.faker.name_female()
        else:
            return self.faker.name()

    def _first_name(self):
        if NamePlaceholderOption.GENDER_MALE in self.option:
            return self.faker.first_name_male()
        elif NamePlaceholderOption.GENDER_FEMALE in self.option:
            return self.faker.first_name_female()
        else:
            return self.faker.first_name()

    def _last_name(self):
        if NamePlaceholderOption.GENDER_MALE in self.option:
            return self.faker.last_name_male()
        elif NamePlaceholderOption.GENDER_FEMALE in self.option:
            return self.faker.last_name_female()
        else:
            return self.faker.last_name()



import re

from jdog.parser import SchemeParser


class Jdog:
    """Proxy class to parser. Accepts language to use, parsing scheme and generating new data

    :var str lang: Language code to use. See :class:`faker.Faker` for supported languages.
    :var boolean strict: If parser should raise NoMatchingPlaceholder error when parsing placeholders.
    """

    def __init__(self, lang='en-US', strict=False):
        """
        :param str lang: Language code to use. See :class:`Faker` for supported languages.
        :param boolean strict: If parser should raise NoMatchingPlaceholder error when parsing placeholders.
        """
        self.parser = SchemeParser(lang, strict)
        self.scheme = None
        self.root = None

    def parse_scheme(self, scheme):
        """
        Parse scheme for generator

        :param str scheme: Scheme to use and parse

        :raises: :class:`.jdog.parser.NoMatchingPlaceholder`, :class:`json.JsonDecodeError`
        """
        self.scheme = scheme
        self.root = self.parser.parse(scheme)

    def generate(self):
        """ Generate new data instance

        :return: Generated JSON
        :rtype: str
        """
        return self.root.exec()

    def add_matcher(self, key, pattern, f_placeholder):
        """ Add or redefine placeholder identified by KEY

                :param str key: Unique placeholder identification.
                :param str pattern: Regex pattern which detects given placeholder.
                :param func f_placeholder: Function which should return :class:`Placeholder` object.
                    Function takes matched token and its arguments - if present.
        """
        self.parser.add_matcher(key, lambda token: re.match(pattern, token), f_placeholder)

    def placeholder_keys(self):
        """ Returns all defined placeholder keys

            :returns: List of placeholder keys
            :rtype: list
        """
        return self.parser.placeholder_keys()



import click
import json
from . import Jdog
from jdog.parser import NoMatchingPlaceholder
from faker.config import AVAILABLE_LOCALES


def _lang_help(ctx, prop, value):
    if not value:
        return
    click.echo('Use one of the following language code: ')
    click.echo(AVAILABLE_LOCALES)
    ctx.exit()


def _validate_lang(ctx, prop, value):
    if value not in AVAILABLE_LOCALES:
        raise click.BadParameter('Language code is invalid. See --lang-help for accepted values')

    return value


def _print_output(result, pretty, output):
    if pretty:
        pretty_output = json.dumps(json.loads(result), indent=4)
        if output:
            output.write(pretty_output)
        else:
            print(pretty_output)
    else:
        if output:
            output.write(result)
        else:
            click.echo(result)


# developer note: Although we cane use for -l parameter click.Choice,
# there is not way to truncate / or hide all options and its look ugly
# so I have decided to introduce special option --lang-help to show and do manual validation for -l option


@click.command('jdog')
@click.argument('scheme', type=click.File('r'))
@click.option('-p', '--pretty', is_flag=True, default=False, help='Output as pretty JSON.')
@click.option('-s', '--strict', is_flag=True, default=False, help='Raise error when no matching placeholder is found.')
@click.option('-l', '--lang', default='en_US', help='Language to use.', callback=_validate_lang)
@click.option('--lang-help', is_flag=True, default=False, help='Displays available language codes and exit.',
              callback=_lang_help)
@click.option('-o', '--output', type=click.File('w'), help='Output file where result is written.')
def run(scheme, strict, pretty, lang, lang_help, output):
    """Accepts SCHEME and generates new data to stdin or to specified OUTPUT"""
    try:
        jdog = Jdog(lang, strict)
        scheme_text = scheme.read()
        jdog.parse_scheme(scheme_text)
        result = jdog.generate()
        _print_output(result, pretty, output)
    except json.JSONDecodeError as e:
        raise click.UsageError(f'Provided SCHEME does not have valid JSON format.\n\tJsonError: {e}')
    except NoMatchingPlaceholder as e:
        raise click.UsageError(e)



import random


class _GroupNode:
    """ Base class for nodes which groups other nodes together.

        :var str begin_token: char to print at the start of group
        :var str end_token: char to print at the end of group
        :var list nodes: Nodes to recursively included within group
    """
    def __init__(self, begin_token, end_token, nodes):
        self.begin_token = begin_token
        self.end_token = end_token
        self.nodes = nodes

    def exec(self):
        """Join nodes together between :attr:`begin_token` and :attr:`end_token`"""
        s = ','.join(str(n.exec()) for n in self.nodes)
        return f'{self.begin_token}{s}{self.end_token}'


class ObjectNode(_GroupNode):
    """
        Represents JSON-Object

        :var list properties: Properties to include within object.
    """
    def __init__(self, properties):
        super().__init__('{', '}', properties)


class ArrayNode(_GroupNode):
    """
        Represents JSON-Array

        :var list properties: Values to include within object.
    """
    def __init__(self, children):
        super().__init__('[', ']', children)


class FuncNode:
    """
        Node accepting function which will be executed later

        :var func f: Function to execute. Has no arguments.
    """
    def __init__(self, f):
        self.f = f

    def exec(self):
        """Executes :attr:`f`"""
        return self.f()


class RangeNode:
    """
        Node which repeatedly creates sub-nodes. Combination of property with array value.

        :var str name: Represents property name which will be generated.
        :var int l: How many times generation runs.
        :var Node child: Object to add to the generated array.
        :var int h: Optional. If specified, generation runs randomly between l (inlcusive) to h (exclusive).
    """
    def __init__(self, name, l, child, h=None):
        self.name = name
        self.low = int(l)
        self.high = int(h) if h else None
        self.child = child

    def exec(self):
        """
            :return: Range representation: JSON property named by :attr:`value` with array value.
            :rtype: str
        """
        children = []
        if self.high is None:
            for i in range(self.low):
                children.append(self.child)
        else:
            for i in range(random.randint(self.low, self.high - 1)):
                children.append(self.child)

        return f'"{self.name}": {ArrayNode(children).exec()}'


class PlaceholderNode:
    """
        Represents node with object

        :var object placeholder: :class:`~jdog.placeholder.placeholder.Placeholder` object to use.
    """
    def __init__(self, placeholder):
        self.placeholder = placeholder

    def exec(self):
        """
            Executes placeholder

        :return: Value returned by placeholder
        """
        return self.placeholder.exec()


class PropertyNode:
    """
        Represent JSON property with name and value

        :var Node name: Property name
        :var Node child: Property value
    """
    def __init__(self, name, child):
        self.name = name
        self.child = child

    def exec(self):
        """
            Executes name node and child if present.
            :return: Property representation
        """
        if self.child:
            return f'{self.name.exec()}:{self.child.exec()}'
        else:
            return f'{self.name.exec()}:""'


class ScalarNode:
    """
        Represents scalar value - that is any number or any arbitrary string

        :var str|number value: Scalar value
    """
    def __init__(self, value):
        self.value = value

    def exec(self):
        """
            If value is string - return as JSON string value otherwise as is
            :return: Value enclosed as a string or as is
        """
        if isinstance(self.value, str):
            return f'"{self.value}"'
        else:
            return self.value



from setuptools import setup, find_packages


with open('README.rst', encoding='utf-8') as f:
    long_description = ''.join(f.readlines())


setup(
    name='jdog',
    version='1.0.0',
    description='Just another Data Offline Generator',
    long_description=long_description,
    author='Petr Nymsa',
    author_email='nymsa.petr@outlook.cz',
    keywords='data generator, data generation, json scheme, scheme based, cli tool',
    license='MIT',
    url='https://github.com/petrnymsa/jdog',
    packages=find_packages(),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Environment :: Console',
        'Topic :: Software Development :: Libraries',
    ],
    entry_points={
        'console_scripts': [
            'jdog = jdog.cli:run',
        ],
    },
    install_requires=['Faker', 'click>=6'],
    zip_safe=False,
)



from jdog.placeholder.placeholder import Placeholder


class FakerPlaceholder(Placeholder):
    """
        Placeholder for faker user

        :var faker.Faker faker: Faker instance
    """
    def __init__(self, full_name, arguments, faker):
        super().__init__(full_name, arguments)
        self.faker = faker
